# AsciiArtWidget v 0.3 by Chris Allen <fastjack@turf.org>
# Copyright Chris Allen 1999
# Last modified Aug 31, 1999
# This software is released under the GPL v2.0 or at your option
# any later version

# API functions:
#    new <frame> <width> <height> : creates a new widget and returns it
#    load <widget> <data> : load data into widget destructively

# TODO:
#    Perhaps more complex editing functions and bindings
#    Testing

# Changes:
# 0.3 -
#	Fixed bug in block-replace
#
# 0.2 -
#	Added block select instead of linear
#	Added block delete instead of linear
#

package provide AsciiArtWidget 0.3

namespace eval ::AsciiArtWidget {
array set info {}
proc new {frame width length} {
	variable info
	set info($frame) { }
	set info($frame,length) $length
	set info($frame,width) $width
	if {$::tcl_platform(platform) == "windows"} {
		set font fixedsys
	} {
		set font fixed
	}
	text $frame -width $width -height $length -font $font -wrap none
	bind $frame <B1-Motion> {
		if {[%W get @%x,%y] == "\n"} {break}
		%W delete @%x,%y
		%W insert @%x,%y $::AsciiArtWidget::info(%W)
		tkTextButton1 %W %x %y
		break
	}
	bind $frame <B3-Motion> {
		if {[%W get @%x,%y] == "\n"} {break}
		%W delete @%x,%y
		%W insert @%x,%y { }
		tkTextButton1 %W %x %y
		break
	}
	bind $frame <Shift-1> {
		%W tag remove sel 1.0 end
		%W mark set insert [tkTextClosestGap %W %x %y]
		%W mark set anchor insert
		break
	}
	bind $frame <Shift-B1-Motion> {
		AsciiArtWidget::boxSelect %W %x %y
		break
	}
	bind $frame <Tab> {break}
	bind $frame <Control-i> {break}
	bind $frame <Return> {break}
	bind $frame <Delete> {
		if {[%W tag nextrange sel 1.0 end] != ""} {
			AsciiArtWidget::wipeArea %W [%W index sel.first] [%W index sel.last]
			break
		} else {
			if {[%W get insert] == "\n"} {tkTextSetCursor %W insert+1c}
			%W delete insert
			%W insert insert { }
			break
		}
	}
	bind $frame <BackSpace> {
		if {[%W tag nextrange sel 1.0 end] != ""} {
			AsciiArtWidget::wipeArea %W [%W index sel.first] [%W index sel.last]
			break
		} {
		     if {[%W get insert-1c] == "\n"} {tkTextSetCursor %W insert-1c}
			%W delete insert-1c
			%W insert insert { }
			tkTextSetCursor %W [%W index {insert - 1c}]
			break
		}
	}
	bind $frame <<Cut>> {break}
	bind $frame <<Paste>> {break}
	bind $frame <Insert> {break}
	bind $frame <KeyPress> {
		set t %A
		if {$t == ""} {unset t; break}
		unset t
		if {[%W get insert] == "\n"} {tkTextSetCursor %W insert+1c}
		if [%W compare insert >= "$::AsciiArtWidget::info(%W,length).$::AsciiArtWidget::info(%W,width)"] {break}
		AsciiArtWidget::charinsert %W %A
		set ::AsciiArtWidget::info(%W) %A
		break
	}
	bind $frame <Control-d> {break}
	bind $frame <Control-k> {break}
	bind $frame <Control-t> {break}
	bind $frame <Meta-d> {break}
	bind $frame <Escape> {break}
	bind $frame <Meta-Delete> {
		if {[%W tag nextrange sel 1.0 end] != ""} {
			AsciiArtWidget::wipeArea %W [%W index sel.first] [%W index sel.last]
			break
		} else {
			if {[%W get insert] == "\n"} {tkTextSetCursor %W insert+1c}
			%W delete insert
			%W insert insert { }
			break
		}
	}
	bind $frame <Meta-BackSpace> {
		if {[%W get insert-1c] == "\n"} {tkTextSetCursor %W insert-1c}
		%W delete insert-1c
		%W insert insert { }
		tkTextSetCursor %W [%W index {insert - 1c}]
		break
	}
	bind $frame <Control-h> {break}
	bind $frame <Left> {
		tkTextSetCursor %W insert-1c
		break
	}
	bind $frame <Shift-Left> {
		tkTextKeySelect %W [%W index {insert - 1c}]
		break
	}
	bind $frame <Shift-Right> {
		tkTextKeySelect %W [%W index {insert + 1c}]
		break
	}
	bind $frame <Shift-Up> {
		tkTextKeySelect %W [tkTextUpDownLine %W -1]
		break
	}
	bind $frame <Shift-Down> {
		tkTextKeySelect %W [tkTextUpDownLine %W 1]
		break
	}
	bind $frame <Right> {
		tkTextSetCursor %W insert+1c
		break
	}
	bind $frame <Up> {
		tkTextSetCursor %W [tkTextUpDownLine %W -1]
		break
	}
	bind $frame <Down> {
		tkTextSetCursor %W [tkTextUpDownLine %W 1]
		break
	}
	bind $frame <B1-Leave> {break}
	bind $frame <Double-1> {break}
	bind $frame <Triple-1> {break}
	bind $frame <Double-Shift-1> {break}
	bind $frame <Triple-Shift-1> {break}
	bind $frame <Shift-Left> {break}
	bind $frame <Shift-Right> {break}
	bind $frame <Shift-Up> {break}
	bind $frame <Shift-Down> {break}
	
	# Twice... don't ask, just leave it.
	wipeArea $frame 1.0 $length.$width
	wipeArea $frame 1.0 $length.$width
	$frame delete end-1c
	$frame config -cursor left_ptr
	return $frame
}
proc charinsert {w s} {
	if {($s == "") || ([$w cget -state] == "disabled")} {
	return
	}
	catch {
		if {[$w compare sel.first <= insert]
			&& [$w compare sel.last >= insert]} {
			wipeArea $w [$w index sel.first] [$w index sel.last]
			$w mark set insert insert-1c
		}
	}
	$w delete insert	
	$w insert insert $s
	$w see insert
}
proc wipeArea {widget start end} {
	regexp {([0-9]+)\.([0-9]+)} $start junk sy sx
	regexp {([0-9]+)\.([0-9]+)} $end junk ey ex
	set ox $sx
	if {$ey > $::AsciiArtWidget::info($widget,length)} {
		set ey $::AsciiArtWidget::info($widget,length)
		set ex $::AsciiArtWidget::info($widget,width)
	}
	for {} {$sy <= $ey} {incr sy} {
		for {} {$sx < $ex} {incr sx} {
			if {[$widget get $sy.$sx] != "\n"} {
				$widget delete $sy.$sx
			}
			$widget insert $sy.$sx { }
		}
		if {[$widget get $sy.$sx] != "\n" && $sy != $ey && $sx == $::AsciiArtWidget::info($widget,width)} {
			$widget insert $sy.$sx \n
		}
		set sx $ox
	}
}
proc boxSelect {widget x y} {
	set cur [tkTextClosestGap $widget $x $y]
	if {[catch {$widget index anchor}]} {
		$widget mark set anchor $cur
	}
	set anchor [$widget index anchor]
	if [$widget compare $cur == $anchor] {return}
	$widget tag remove sel 1.0 end
	set cur [split $cur {. }]
	set anchor [split $anchor .]
	lappend ys [lindex $cur 0] [lindex $anchor 0]
	lappend xs [lindex $cur 1] [lindex $anchor 1]
	set ys [lsort -integer $ys]
	set xs [lsort -integer $xs]
	set minx [lindex $xs 0]
	set maxx [lindex $xs 1]
	set miny [lindex $ys 0]
	set maxy [lindex $ys 1]
	for {} {$miny <= $maxy} {incr miny} {
		for {set ox $minx} {$ox <= $maxx} {incr ox} {
			if {[$widget get $miny.$ox] == "\n"} {continue}
			$widget tag add sel $miny.$ox
		}
	}
}
proc load {widget data} {
	if [regexp \n $data] {
		regsub -all \r $data {} data
	} {
		regsub -all \r $data \n data
	}
	$widget delete 1.0 end
	$widget insert end $data
	set height [lindex [split [$widget index insert] { .}] 0]
	set ::AsciiArtWidget::info($widget,length) $height
	set width 0
	foreach of [split $data \n] {
		set slen [string length $of]
		if {$slen > $width} {set width $slen}
	}
	set ::AsciiArtWidget::info($widget,width) $width
	$widget config -height $height -width $width
	square $widget $width $height
}
proc square {widget width height} {
	for {set line 1} {$line <= $height} {incr line} {
		set data [$widget get $line.0 $line.end]
		set pad [expr $width - [string length $data]]
		for {set i 0} {$i < $pad} {incr i} {
			append data { }
		}
		$widget delete $line.0 $line.end
		$widget insert $line.0 $data
	}	
}
}
